"""
Tento program:
- Načte data o publikacích ze vstupu (rok, časopis, název, autoři)
- Vytřídí pouze publikace od roku 2015 a novější
- Vypíše počet publikací ke každému autorovi (abecedně)
- TODO: Vypíše autora s největším počtem publikací

Spuštění programu:
- Z příkazového řádku:
cat data.txt | python papers.py
(Tento zápis znamená, že nemusíme vstup psát ručně, ale načte se ze souboru data.txt)

Úkoly:
1. Zhezčit zápis programu (stylistická doporučení (např. mezery před a za =, pojmenování proměnných)
2. Zpřehlednit program rozdělením do funkcí
3. Zefektivnit program tak, aby běžel rychle i na velkém datovém souboru (data_big.txt)
4. Zjednodušit program pomocí rozbalení, generátorových výrazů, případně funkcí enumerate, sorted...
5. Doplnit poslední krok (výběr autora s největším počtem publikací)

Rozšiřující učivo:
6. Zpřehlednit program použitím NamedTuple
7. Upravit program použitím walrus operator (:=), generátorové funkce, lambda funkce (zde nutně nezefektivní ani nezpřehlední, spíš věc vkusu)
"""

from __future__ import annotations
from typing import NamedTuple, Iterator


class Paper(NamedTuple):
    year: int
    journal: str
    title: str
    authors: list[str]


def log(name: str, items: list | dict) -> None:
    """Print length of a list/dict and its first item"""
    print(f'{name}: {len(items)}')
    if len(items) > 0:
        if isinstance(items, dict):
            for key, value in items.items():
                print(f'{key}: {value}')
                break
        else:
            print(items[0])
    print()


def read_lines() -> Iterator[str]:
    """Read lines from input until the first empty line"""
    while (line:= input()) != '':
        yield line


def parse_paper(line: str) -> Paper:
    """Parse line as a paper in this format: year;journal;title;authors..."""
    year, journal, title, *authors = line.split(';')
    return Paper(int(year), journal, title, authors)


def get_authors(papers: list[Paper]) -> list[str]:
    """Return the list of all authors of papers (in alphabetical order)."""
    author_set = {author for paper in papers for author in paper.authors}
    return sorted(author_set)


def sort_paper_by_author(papers: list[Paper]) -> dict[str, list[Paper]]:
    """Return the dict mapping authors to list of their papers."""
    authors = get_authors(papers)
    papers_per_author = {author: [] for author in authors}
    for paper in papers:
        for author in paper.authors:
            papers_per_author[author].append(paper)
    return papers_per_author


def print_overview(papers_per_author: dict[str, list[Paper]]) -> None:
    """For each author, print their name and number of papers."""
    print('Overview of number of papers per author:')
    for author, papers in papers_per_author.items():
        print(f'    {author}: {len(papers)}')
    print()


def get_top_author(papers_per_author: dict[str, list[Paper]]) -> str:
    """Get the author with the highest number published papers."""
    return max(papers_per_author.keys(), key = lambda a: len(papers_per_author[a]))


def main():
    all_papers = [parse_paper(line) for line in read_lines()]
    # or: all_papers = list(map(parse_paper, read_lines()))
    log('Parsed papers', all_papers)

    filtered_papers = [paper for paper in all_papers if paper.year >= 2015]
    # or: filtered_papers = list(filter(lambda p: p.year >= 2015, all_papers))
    log('Filtered papers', filtered_papers)

    all_authors = get_authors(filtered_papers)
    log('Authors', all_authors) 

    papers_per_author = sort_paper_by_author(filtered_papers)
    log('Assigned papers to authors', papers_per_author)

    print_overview(papers_per_author)

    print('The author with most papers:')
    print(f'    {get_top_author(papers_per_author)}')
    print()


main()

